@@ -48,6 +48,10 @@
 
 #include <libnetfilter_conntrack/libnetfilter_conntrack.h>
 
+#ifndef NSEC_PER_SEC
+#define NSEC_PER_SEC    1000000000L
+#endif
+
 typedef enum TIMES_ { START, STOP, __TIME_MAX } TIMES;
 
 struct ct_timestamp {
@@ -177,6 +181,8 @@ enum nfct_keys {
 	NFCT_REPLY_RAW_PKTCOUNT,
 	NFCT_ICMP_CODE,
 	NFCT_ICMP_TYPE,
+	NFCT_ICMPV6_CODE,
+	NFCT_ICMPV6_TYPE,
 	NFCT_CT_MARK,
 	NFCT_CT_ID,
 	NFCT_CT_EVENT,
@@ -236,7 +242,7 @@ static struct ulogd_key nfct_okeys[] = {
 		},
 	},
 	{
-		.type	= ULOGD_RET_UINT32,
+		.type	= ULOGD_RET_UINT64,
 		.flags	= ULOGD_RETF_NONE,
 		.name	= "orig.raw.pktlen",
 		.ipfix	= {
@@ -246,7 +252,7 @@ static struct ulogd_key nfct_okeys[] = {
 		},
 	},
 	{
-		.type	= ULOGD_RET_UINT32,
+		.type	= ULOGD_RET_UINT64,
 		.flags	= ULOGD_RETF_NONE,
 		.name	= "orig.raw.pktcount",
 		.ipfix	= {
@@ -301,7 +307,7 @@ static struct ulogd_key nfct_okeys[] = {
 		},
 	},
 	{
-		.type	= ULOGD_RET_UINT32,
+		.type	= ULOGD_RET_UINT64,
 		.flags	= ULOGD_RETF_NONE,
 		.name	= "reply.raw.pktlen",
 		.ipfix	= {
@@ -311,7 +317,7 @@ static struct ulogd_key nfct_okeys[] = {
 		},
 	},
 	{
-		.type	= ULOGD_RET_UINT32,
+		.type	= ULOGD_RET_UINT64,
 		.flags	= ULOGD_RETF_NONE,
 		.name	= "reply.raw.pktcount",
 		.ipfix	= {
@@ -338,6 +344,24 @@ static struct ulogd_key nfct_okeys[] = {
 			.field_id	= IPFIX_icmpTypeIPv4,
 		},
 	},
+	{
+		.type	= ULOGD_RET_UINT8,
+		.flags	= ULOGD_RETF_NONE,
+		.name	= "icmpv6.code",
+		.ipfix	= {
+			.vendor		= IPFIX_VENDOR_IETF,
+			.field_id	= IPFIX_icmpCodeIPv6,
+		},
+	},
+	{
+		.type	= ULOGD_RET_UINT8,
+		.flags	= ULOGD_RETF_NONE,
+		.name	= "icmpv6.type",
+		.ipfix	= {
+			.vendor		= IPFIX_VENDOR_IETF,
+			.field_id	= IPFIX_icmpTypeIPv6,
+		},
+	},
 	{
 		.type	= ULOGD_RET_UINT32,
 		.flags	= ULOGD_RETF_NONE,
@@ -543,6 +567,12 @@ static int propagate_ct(struct ulogd_pluginstance *main_upi,
 		okey_set_u16(&ret[NFCT_ICMP_TYPE],
 			     nfct_get_attr_u8(ct, ATTR_ICMP_TYPE));
 		break;
+	case IPPROTO_ICMPV6:
+		okey_set_u16(&ret[NFCT_ICMPV6_CODE],
+			     nfct_get_attr_u8(ct, ATTR_ICMP_CODE));
+		okey_set_u16(&ret[NFCT_ICMPV6_TYPE],
+			     nfct_get_attr_u8(ct, ATTR_ICMP_TYPE));
+		break;
 	}
 
 	switch (nfct_get_attr_u8(ct, ATTR_REPL_L4PROTO)) {
@@ -557,14 +587,14 @@ static int propagate_ct(struct ulogd_pluginstance *main_upi,
 			     htons(nfct_get_attr_u16(ct, ATTR_REPL_PORT_DST)));
 	}
 
-	okey_set_u32(&ret[NFCT_ORIG_RAW_PKTLEN],
-		     nfct_get_attr_u32(ct, ATTR_ORIG_COUNTER_BYTES));
-	okey_set_u32(&ret[NFCT_ORIG_RAW_PKTCOUNT],
-		     nfct_get_attr_u32(ct, ATTR_ORIG_COUNTER_PACKETS));
-	okey_set_u32(&ret[NFCT_REPLY_RAW_PKTLEN],
-		     nfct_get_attr_u32(ct, ATTR_REPL_COUNTER_BYTES));
-	okey_set_u32(&ret[NFCT_REPLY_RAW_PKTCOUNT],
-		     nfct_get_attr_u32(ct, ATTR_REPL_COUNTER_PACKETS));
+	okey_set_u64(&ret[NFCT_ORIG_RAW_PKTLEN],
+		     nfct_get_attr_u64(ct, ATTR_ORIG_COUNTER_BYTES));
+	okey_set_u64(&ret[NFCT_ORIG_RAW_PKTCOUNT],
+		     nfct_get_attr_u64(ct, ATTR_ORIG_COUNTER_PACKETS));
+	okey_set_u64(&ret[NFCT_REPLY_RAW_PKTLEN],
+		     nfct_get_attr_u64(ct, ATTR_REPL_COUNTER_BYTES));
+	okey_set_u64(&ret[NFCT_REPLY_RAW_PKTCOUNT],
+		     nfct_get_attr_u64(ct, ATTR_REPL_COUNTER_PACKETS));
 
 	okey_set_u32(&ret[NFCT_CT_MARK], nfct_get_attr_u32(ct, ATTR_MARK));
 	okey_set_u32(&ret[NFCT_CT_ID], nfct_get_attr_u32(ct, ATTR_ID));
@@ -617,6 +647,33 @@ do_propagate_ct(struct ulogd_pluginstance *upi,
 	propagate_ct(upi, upi, ct, type, ts);
 }
 
+static int set_timestamp_from_ct_try(struct ct_timestamp *ts,
+				   struct nf_conntrack *ct, int name)
+{
+	int attr_name;
+
+	if (name == START)
+		attr_name = ATTR_TIMESTAMP_START;
+	else
+		attr_name = ATTR_TIMESTAMP_STOP;
+
+	if (nfct_attr_is_set(ct, attr_name)) {
+		ts->time[name].tv_sec =
+		     nfct_get_attr_u64(ct, attr_name) / NSEC_PER_SEC;
+		ts->time[name].tv_usec =
+		     (nfct_get_attr_u64(ct, attr_name) % NSEC_PER_SEC) / 1000;
+		return 1;
+	}
+	return 0;
+}
+
+static void set_timestamp_from_ct(struct ct_timestamp *ts,
+				   struct nf_conntrack *ct, int name)
+{
+	if (!set_timestamp_from_ct_try(ts, ct, name))
+		gettimeofday(&ts->time[name], NULL);
+}
+
 static int
 event_handler_hashtable(enum nf_conntrack_msg_type type,
 			struct nf_conntrack *ct, void *data)
@@ -629,13 +686,13 @@ event_handler_hashtable(enum nf_conntrack_msg_type type,
 
 	switch(type) {
 	case NFCT_T_NEW:
-		ts = calloc(sizeof(struct ct_timestamp), 1);
+		ts = calloc(1, sizeof(struct ct_timestamp));
 		if (ts == NULL)
 			return NFCT_CB_CONTINUE;
 
 		ts->ct = ct;
-		gettimeofday(&ts->time[START], NULL);
 
+		set_timestamp_from_ct(ts, ct, START);
 		id = hashtable_hash(cpi->ct_active, ct);
 		ret = hashtable_add(cpi->ct_active, &ts->hashnode, id);
 		if (ret < 0) {
@@ -650,13 +707,12 @@ event_handler_hashtable(enum nf_conntrack_msg_type type,
 		if (ts)
 			nfct_copy(ts->ct, ct, NFCT_CP_META);
 		else {
-			ts = calloc(sizeof(struct ct_timestamp), 1);
+			ts = calloc(1, sizeof(struct ct_timestamp));
 			if (ts == NULL)
 				return NFCT_CB_CONTINUE;
 
 			ts->ct = ct;
-			gettimeofday(&ts->time[START], NULL);
-
+			set_timestamp_from_ct(ts, ct, START);
 			ret = hashtable_add(cpi->ct_active, &ts->hashnode, id);
 			if (ret < 0) {
 				free(ts);
@@ -670,7 +726,7 @@ event_handler_hashtable(enum nf_conntrack_msg_type type,
 		ts = (struct ct_timestamp *)
 			hashtable_find(cpi->ct_active, ct, id);
 		if (ts) {
-			gettimeofday(&ts->time[STOP], NULL);
+			set_timestamp_from_ct(ts, ct, STOP);
 			do_propagate_ct(upi, ct, type, ts);
 			hashtable_del(cpi->ct_active, &ts->hashnode);
 			nfct_destroy(ts->ct);
@@ -679,7 +735,7 @@ event_handler_hashtable(enum nf_conntrack_msg_type type,
 			struct ct_timestamp tmp = {
 				.ct = ct,
 			};
-			gettimeofday(&tmp.time[STOP], NULL);
+			set_timestamp_from_ct(&tmp, ct, STOP);
 			tmp.time[START].tv_sec = 0;
 			tmp.time[START].tv_usec = 0;
 			do_propagate_ct(upi, ct, type, &tmp);
@@ -704,14 +760,16 @@ event_handler_no_hashtable(enum nf_conntrack_msg_type type,
 
 	switch(type) {
 	case NFCT_T_NEW:
-		gettimeofday(&tmp.time[START], NULL);
+		set_timestamp_from_ct(&tmp, ct, START);
 		tmp.time[STOP].tv_sec = 0;
 		tmp.time[STOP].tv_usec = 0;
 		break;
 	case NFCT_T_DESTROY:
-		gettimeofday(&tmp.time[STOP], NULL);
-		tmp.time[START].tv_sec = 0;
-		tmp.time[START].tv_usec = 0;
+		set_timestamp_from_ct(&tmp, ct, STOP);
+		if (!set_timestamp_from_ct_try(&tmp, ct, START)) {
+			tmp.time[START].tv_sec = 0;
+			tmp.time[START].tv_usec = 0;
+		}
 		break;
 	default:
 		ulogd_log(ULOGD_NOTICE, "unsupported message type\n");
@@ -739,12 +797,12 @@ polling_handler(enum nf_conntrack_msg_type type,
 		if (ts)
 			nfct_copy(ts->ct, ct, NFCT_CP_META);
 		else {
-			ts = calloc(sizeof(struct ct_timestamp), 1);
+			ts = calloc(1, sizeof(struct ct_timestamp));
 			if (ts == NULL)
 				return NFCT_CB_CONTINUE;
 
 			ts->ct = ct;
-			gettimeofday(&ts->time[START], NULL);
+			set_timestamp_from_ct(ts, ct, START);
 
 			ret = hashtable_add(cpi->ct_active, &ts->hashnode, id);
 			if (ret < 0) {
@@ -853,6 +911,7 @@ static int do_purge(void *data1, void *data2)
 	/* if it is not in kernel anymore, purge it */
 	ret = nfct_query(cpi->pgh, NFCT_Q_GET, ts->ct);
 	if (ret == -1 && errno == ENOENT) {
+		set_timestamp_from_ct(ts, ts->ct, STOP);
 		do_propagate_ct(upi, ts->ct, NFCT_T_DESTROY, ts);
 		hashtable_del(cpi->ct_active, &ts->hashnode);
 		nfct_destroy(ts->ct);
@@ -876,12 +935,12 @@ static int overrun_handler(enum nf_conntrack_msg_type type,
 	ts = (struct ct_timestamp *)
 		hashtable_find(cpi->ct_active, ct, id);
 	if (ts == NULL) {
-		ts = calloc(sizeof(struct ct_timestamp), 1);
+		ts = calloc(1, sizeof(struct ct_timestamp));
 		if (ts == NULL)
 			return NFCT_CB_CONTINUE;
 
 		ts->ct = ct;
-		gettimeofday(&ts->time[START], NULL); /* do our best here */
+		set_timestamp_from_ct(ts, ct, START);
 
 		ret = hashtable_add(cpi->ct_active, &ts->hashnode, id);
 		if (ret < 0) {
@@ -931,6 +990,9 @@ dump_reset_handler(enum nf_conntrack_msg_type type,
 	int ret = NFCT_CB_CONTINUE, rc, id;
 	struct ct_timestamp *ts;
 
+	if (!cpi->ct_active)
+		return NFCT_CB_STOP;
+
 	switch(type) {
 	case NFCT_T_UPDATE:
 		id = hashtable_hash(cpi->ct_active, ct);
@@ -939,12 +1001,12 @@ dump_reset_handler(enum nf_conntrack_msg_type type,
 		if (ts)
 			nfct_copy(ts->ct, ct, NFCT_CP_META);
 		else {
-			ts = calloc(sizeof(struct ct_timestamp), 1);
+			ts = calloc(1, sizeof(struct ct_timestamp));
 			if (ts == NULL)
 				return NFCT_CB_CONTINUE;
 
 			ts->ct = ct;
-			gettimeofday(&ts->time[START], NULL);
+			set_timestamp_from_ct(ts, ct, START);
 
 			rc = hashtable_add(cpi->ct_active, &ts->hashnode, id);
 			if (rc < 0) {
@@ -991,18 +1053,6 @@ static void polling_timer_cb(struct ulogd_timer *t, void *data)
 	ulogd_add_timer(&cpi->timer, pollint_ce(upi->config_kset).u.value);
 }
 
-static int configure_nfct(struct ulogd_pluginstance *upi,
-			  struct ulogd_pluginstance_stack *stack)
-{
-	int ret;
-
-	ret = config_parse_file(upi->id, upi->config_kset);
-	if (ret < 0)
-		return ret;
-
-	return 0;
-}
-
 static void overrun_timeout(struct ulogd_timer *a, void *data)
 {
 	int family = AF_UNSPEC;
@@ -1295,7 +1345,7 @@ static int constructor_nfct_events(struct ulogd_pluginstance *upi)
 		setsockopt(nfct_fd(cpi->cth), SOL_NETLINK,
 				NETLINK_NO_ENOBUFS, &on, sizeof(int));
 		ulogd_log(ULOGD_NOTICE, "NFCT reliable logging "
-					"has been enabled.");
+					"has been enabled.\n");
 	}
 	cpi->nfct_fd.fd = nfct_fd(cpi->cth);
 	cpi->nfct_fd.cb = &read_cb_nfct;
@@ -1519,7 +1569,6 @@ static struct ulogd_plugin nfct_plugin = {
 	},
 	.config_kset 	= &nfct_kset,
 	.interp 	= NULL,
-	.configure	= &configure_nfct,
 	.start		= &constructor_nfct,
 	.stop		= &destructor_nfct,
 	.signal		= &signal_nfct,
