@@ -4,6 +4,7 @@
  *
  * (C) 2000-2005 by Harald Welte <laforge@gnumonks.org>
  * (C) 2013 by Eric Leblond <eric@regit.org>
+ * (C) 2013 Chris Boot <bootc@bootc.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 
@@ -55,12 +56,16 @@
 #include <signal.h>
 #include <dlfcn.h>
 #include <sys/types.h>
+#include <fcntl.h>
 #include <dirent.h>
 #include <getopt.h>
 #include <pwd.h>
 #include <grp.h>
 #include <syslog.h>
 #include <sys/time.h>
+#include <sys/stat.h>
+#include <sched.h>
+#include <limits.h>
 #include <ulogd/conffile.h>
 #include <ulogd/ulogd.h>
 #ifdef DEBUG
@@ -78,11 +83,37 @@
 static FILE *logfile = NULL;		/* logfile pointer */
 static char *ulogd_logfile = NULL;
 static const char *ulogd_configfile = ULOGD_CONFIGFILE;
-static FILE syslog_dummy;
+static const char *ulogd_pidfile = NULL;
+static int ulogd_pidfile_fd = -1;
+static FILE *syslog_dummy;
 
 static int info_mode = 0;
 
 static int verbose = 0;
+static int created_pidfile = 0;
+
+/*
+ * Using self-pipe trick to handle signals safely.
+ * (To avoid calling Async-Signal-Unsafe functions in signal handler.)
+ *
+ * cf. https://lwn.net/Articles/177897/
+ *
+ * Capacity of pipe(2) buffer is >= 4096.
+ */
+static int signal_channel[2] = { -1, -1 };	/* initialize with invalid fd */
+#define	SIGNAL_RX_FD	(signal_channel[0])	/* main thread select() this fd */
+#define	SIGNAL_TX_FD	(signal_channel[1])	/* signal handler writes to this fd */
+
+static void signal_handler(int signal);
+static void sigterm_handler_task(int signal);
+static void signal_handler_task(int signal);
+
+static int  signal_channel_callback(int fd, unsigned int what, void *data);
+static struct ulogd_fd	signal_channel_ulogfd = {
+	.when = ULOGD_FD_READ,
+	.cb = &signal_channel_callback,
+	.data = &signal_channel_ulogfd,
+};
 
 /* linked list for all registered plugins */
 static LLIST_HEAD(ulogd_plugins);
@@ -94,6 +125,7 @@ static LLIST_HEAD(ulogd_pi_stacks);
 static int load_plugin(const char *file);
 static int create_stack(const char *file);
 static int logfile_open(const char *name);
+static void cleanup_pidfile();
 
 static struct config_keyset ulogd_kset = {
 	.num_ces = 4,
@@ -219,6 +251,65 @@ int ulogd_wildcard_inputkeys(struct ulogd_pluginstance *upi)
 	return 0;
 }
 
+/**
+ * Parse the given section in the config file into the given keyset.
+ * Returns ULOGD_IRET_OK on success, ULOGD_IRET_ERR on error.
+ * If an error occurs, writes a descriptive message to the log.
+ */
+int ulogd_parse_configfile(const char *section, struct config_keyset *ce)
+{
+	int err;
+
+	err = config_parse_file(section, ce);
+
+	switch(err) {
+	case 0:
+		return ULOGD_IRET_OK;
+		break;
+	case -ERROPEN:
+		ulogd_log(ULOGD_ERROR, "unable to open configfile: %s\n",
+			  ulogd_configfile);
+		break;
+	case -ERRMAND:
+		ulogd_log(ULOGD_ERROR, "mandatory option \"%s\" not found\n",
+			  config_errce->key);
+		break;
+	case -ERRMULT:
+		ulogd_log(ULOGD_ERROR, "option \"%s\" occurred more than once\n",
+		          config_errce->key);
+		break;
+	case -ERRUNKN:
+		ulogd_log(ULOGD_ERROR, "unknown config key \"%s\"\n",
+		          config_errce->key);
+		free(config_errce);
+		config_errce = NULL;
+		break;
+	case -ERRSECTION:
+		ulogd_log(ULOGD_ERROR, "section \"%s\" not found\n", section);
+		break;
+	case -ERRTOOLONG:
+		if (config_errce != NULL)
+			ulogd_log(ULOGD_ERROR,
+			          "string value too long for key \"%s\"\n",
+			          config_errce->key);
+		else
+			ulogd_log(ULOGD_ERROR, "string value is too long\n");
+		break;
+	case -ERRINTFORMAT:
+		ulogd_log(ULOGD_ERROR,
+		          "integer has invalid format for key \"%s\"\n",
+		          config_errce->key);
+		break;
+	case -ERRINTRANGE:
+		ulogd_log(ULOGD_ERROR,
+		          "integer is out of range for key \"%s\"\n",
+		          config_errce->key);
+		break;
+	}
+
+	return ULOGD_IRET_ERR;
+}
+
 
 /***********************************************************************
  * PLUGIN MANAGEMENT 
@@ -362,6 +453,7 @@ void ulogd_register_plugin(struct ulogd_plugin *me)
 	if (strcmp(me->version, VERSION)) { 
 		ulogd_log(ULOGD_NOTICE, 
 			  "plugin `%s' has incompatible version %s\n",
+			  me->name,
 			  me->version);
 		return;
 	}
@@ -379,6 +471,80 @@ void ulogd_register_plugin(struct ulogd_plugin *me)
 	}
 }
 
+/***********************************************************************
+ * UTILITY FUNCTIONS FOR SIGNAL HANDLING
+ ***********************************************************************/
+static void call_signal_handler_tasks(int sig)
+{
+	/* Deliver signals */
+	switch (sig) {
+	case SIGTERM:
+	case SIGINT:
+		sigterm_handler_task(sig);
+		break;
+	case SIGHUP:
+	case SIGALRM:
+	case SIGUSR1:
+	case SIGUSR2:
+		signal_handler_task(sig);
+		break;
+	default:
+		break;
+	}
+}
+
+static int signal_channel_callback(int fd, unsigned int what, void *data)
+{
+	int	ret;
+	unsigned char c;
+
+	while ((ret = read(fd, &c, 1)) > 0) {
+		call_signal_handler_tasks((int)c);
+	}
+	return 0;
+}
+
+static int create_signal_channel(void)
+{
+	int	ret;
+	long	flags;
+
+	if ((ret = pipe(signal_channel)) < 0)
+		return ret;
+
+	/* SIGNAL_RX_FD, channel[0] */
+	if ((flags = fcntl(SIGNAL_RX_FD, F_GETFL)) < 0)
+		goto err_exit;
+	flags |= O_NONBLOCK;
+	if ( fcntl(SIGNAL_RX_FD, F_SETFL, flags) < 0 )
+		goto err_exit;
+
+	/* SIGNAL_TX_FD, channel[1] */
+	if ((flags = fcntl(SIGNAL_TX_FD, F_GETFL)) < 0)
+		goto err_exit;
+	flags |= O_NONBLOCK;
+	if ( fcntl(SIGNAL_TX_FD, F_SETFL, flags) < 0 )
+		goto err_exit;
+
+	/* register SIGNAL_RX_FD to ulogd */
+	signal_channel_ulogfd.fd = SIGNAL_RX_FD;
+	ulogd_register_fd(&signal_channel_ulogfd);
+
+	return 0;
+
+err_exit:
+	if (SIGNAL_RX_FD >= 0) {
+		(void)close(SIGNAL_RX_FD);
+		SIGNAL_RX_FD = -1;
+	}
+	if (SIGNAL_TX_FD >= 0) {
+		(void)close(SIGNAL_TX_FD);
+		SIGNAL_TX_FD = -1;
+	}
+	return -1;
+}
+
+
 /***********************************************************************
  * MAIN PROGRAM
  ***********************************************************************/
@@ -420,7 +586,7 @@ void __ulogd_log(int level, char *file, int line, const char *format, ...)
 	if (level < loglevel_ce.u.value)
 		return;
 
-	if (logfile == &syslog_dummy) {
+	if (logfile == syslog_dummy) {
 		/* FIXME: this omits the 'file' string */
 		va_start(ap, format);
 		vsyslog(ulogd2syslog_level(level), format, ap);
@@ -435,7 +601,7 @@ void __ulogd_log(int level, char *file, int line, const char *format, ...)
 		timestr = ctime(&tm);
 		timestr[strlen(timestr)-1] = '\0';
 		fprintf(outfd, "%s <%1.1d> %s:%d ", timestr, level, file, line);
-		if (verbose)
+		if (verbose && outfd != stderr)
 			fprintf(stderr, "%s <%1.1d> %s:%d ", timestr, level, file, line);
 
 
@@ -445,7 +611,7 @@ void __ulogd_log(int level, char *file, int line, const char *format, ...)
 		/* flush glibc's buffer */
 		fflush(outfd);
 
-		if (verbose) {
+		if (verbose && outfd != stderr) {
 			va_start(ap, format);
 			vfprintf(stderr, format, ap);
 			va_end(ap);
@@ -457,6 +623,8 @@ void __ulogd_log(int level, char *file, int line, const char *format, ...)
 
 static void warn_and_exit(int daemonize)
 {
+	cleanup_pidfile();
+
 	if (!daemonize) {
 		if (logfile && !verbose) {
 			fprintf(stderr, "Fatal error, check logfile \"%s\""
@@ -552,17 +720,16 @@ pluginstance_alloc_init(struct ulogd_plugin *pl, char *pi_id,
 	}
 	size += pl->input.num_keys * sizeof(struct ulogd_key);
 	size += pl->output.num_keys * sizeof(struct ulogd_key);
-	pi = malloc(size);
+	pi = calloc(1, size);
 	if (!pi)
 		return NULL;
 
 	/* initialize */
-	memset(pi, 0, size);
 	INIT_LLIST_HEAD(&pi->list);
 	INIT_LLIST_HEAD(&pi->plist);
 	pi->plugin = pl;
 	pi->stack = stack;
-	memcpy(pi->id, pi_id, sizeof(pi->id));
+	strncpy(pi->id, pi_id, ULOGD_MAX_KEYLEN);
 
 	ptr = (void *)pi + sizeof(*pi);
 
@@ -620,6 +787,41 @@ static int load_plugin(const char *file)
 	return 0;
 }
 
+static int load_all_plugins(void)
+{
+	DIR *d;
+	struct dirent *dent;
+	char path[PATH_MAX];
+
+	d = opendir(ULOGD2_LIBDIR);
+	if (d == NULL) {
+		ulogd_log(ULOGD_ERROR, "load_all_plugins: opendir(%s): %s\n",
+			  ULOGD2_LIBDIR, strerror(errno));
+		return -1;
+	}
+
+	ulogd_log(ULOGD_NOTICE, "loading all plugins at %s\n", ULOGD2_LIBDIR);
+
+	while ((dent = readdir(d)) != NULL) {
+		if (strcmp(dent->d_name, ".") == 0 ||
+		    strcmp(dent->d_name, "..") == 0)
+			continue;
+
+		int len = strlen(dent->d_name);
+		if (len < 3)
+			continue;
+
+		if (strcmp(&dent->d_name[len - 3], ".so") != 0)
+			continue;
+
+		snprintf(path, sizeof(path), "%s/%s", ULOGD2_LIBDIR,
+			 dent->d_name);
+		if (load_plugin(path) != 0)
+			return -1;
+	}
+	return 0;
+}
+
 /* find an output key in a given stack, starting at 'start' */
 static struct ulogd_key *
 find_okey_in_stack(char *name,
@@ -656,19 +858,29 @@ create_stack_resolve_keys(struct ulogd_pluginstance_stack *stack)
 
 	/* pre-configuration pass */
 	llist_for_each_entry_reverse(pi_cur, &stack->list, list) {
+		int ret;
+
 		ulogd_log(ULOGD_DEBUG, "traversing plugin `%s'\n", 
 			  pi_cur->plugin->name);
 		/* call plugin to tell us which keys it requires in
 		 * given configuration */
 		if (pi_cur->plugin->configure) {
-			int ret = pi_cur->plugin->configure(pi_cur, 
-							    stack);
-			if (ret < 0) {
-				ulogd_log(ULOGD_ERROR, "error during "
-					  "configure of plugin %s\n",
-					  pi_cur->plugin->name);
-				return ret;
-			}
+			ret = pi_cur->plugin->configure(pi_cur, stack);
+		} else {
+			struct config_keyset empty_kset = {.num_ces=0};
+			struct config_keyset *kset = &empty_kset;
+
+			if (pi_cur->config_kset)
+				kset = pi_cur->config_kset;
+
+			ret = ulogd_parse_configfile(pi_cur->id, kset);
+		}
+
+		if (ret < 0) {
+			ulogd_log(ULOGD_ERROR, "error during "
+				  "configure of plugin %s\n",
+				  pi_cur->plugin->name);
+			return ret;
 		}
 	}
 
@@ -817,8 +1029,10 @@ static int create_stack(const char *option)
 	char *tok;
 	int ret;
 
+	if (llist_empty(&ulogd_plugins_handle))
+		load_all_plugins();
+
 	if (!buf) {
-		ulogd_log(ULOGD_ERROR, "");
 		ret = -ENOMEM;
 		goto out_buf;
 	}
@@ -830,7 +1044,7 @@ static int create_stack(const char *option)
 	}
 	INIT_LLIST_HEAD(&stack->list);
 
-	ulogd_log(ULOGD_DEBUG, "building new pluginstance stack (%s):\n",
+	ulogd_log(ULOGD_NOTICE, "building new pluginstance stack: '%s'\n",
 		  option);
 
 	/* PASS 1: find and instanciate plugins of stack, link them together */
@@ -916,7 +1130,6 @@ static void ulogd_main_loop(void)
 	struct timeval *next = NULL;
 
 	while (1) {
-		/* XXX: signal blocking? */
 		if (next != NULL && !timerisset(next))
 			next = ulogd_do_timer_run(&next_alarm);
 		else
@@ -941,7 +1154,7 @@ static int logfile_open(const char *name)
 		logfile = stdout;
 	} else if (!strcmp(name, "syslog")) {
 		openlog("ulogd", LOG_PID, LOG_DAEMON);
-		logfile = &syslog_dummy;
+		logfile = syslog_dummy = fopen("/dev/null", "w");
 	} else {
 		logfile = fopen(ulogd_logfile, "a");
 		if (!logfile) {
@@ -954,52 +1167,165 @@ static int logfile_open(const char *name)
 	return 0;
 }
 
-/* wrapper to handle conffile error codes */
-static int parse_conffile(const char *section, struct config_keyset *ce)
+/*
+ * Apply F_WRLCK to fd using fcntl().
+ *
+ * This function is copied verbatim from atd's daemon.c file, published under
+ * the GPL2+ license with the following copyright statement:
+ * Copyright (C) 1996 Thomas Koenig
+ */
+static int lock_fd(int fd, int wait)
 {
-	int err;
+	struct flock lock;
 
-	err = config_parse_file(section, ce);
+	lock.l_type = F_WRLCK;
+	lock.l_whence = SEEK_SET;
+	lock.l_start = 0;
+	lock.l_len = 0;
 
-	switch(err) {
-		case 0:
-			return 0;
-			break;
-		case -ERROPEN:
+	if (wait)
+		return fcntl(fd, F_SETLKW, &lock);
+	else
+		return fcntl(fd, F_SETLK, &lock);
+}
+
+/*
+ * Manage ulogd's pidfile.
+ *
+ * This function is based on atd's daemon.c:daemon_setup() function, published
+ * under the GPL2+ license with the following copyright statement:
+ * Copyright (C) 1996 Thomas Koenig
+ */
+static int create_pidfile()
+{
+	int fd;
+	FILE *fp;
+	pid_t pid = -1;
+
+	if (!ulogd_pidfile)
+		return 0;
+
+	fd = open(ulogd_pidfile, O_RDWR | O_CREAT | O_EXCL, 0644);
+	if (fd < 0) {
+		if (errno != EEXIST) {
+			ulogd_log(ULOGD_ERROR, "cannot open %s: %d\n",
+					ulogd_pidfile, errno);
+			return -1;
+		}
+
+		fd = open(ulogd_pidfile, O_RDWR);
+		if (fd < 0) {
+			ulogd_log(ULOGD_ERROR, "cannot open %s: %d\n",
+					ulogd_pidfile, errno);
+			return -1;
+		}
+
+		fp = fdopen(fd, "rw");
+		if (fp == NULL) {
+			ulogd_log(ULOGD_ERROR, "cannot fdopen %s: %d\n",
+					ulogd_pidfile, errno);
+			close(fd);
+			return -1;
+		}
+
+		if ((fscanf(fp, "%d", &pid) != 1) || (pid == getpid())
+				|| (lock_fd(fd, 0) == 0)) {
+			ulogd_log(ULOGD_NOTICE,
+				  "removing stale pidfile for pid %d\n", pid);
+
+			if (unlink(ulogd_pidfile) < 0) {
+				ulogd_log(ULOGD_ERROR, "cannot unlink %s: %d\n",
+						ulogd_pidfile, errno);
+				return -1;
+			}
+		} else {
+			ulogd_log(ULOGD_FATAL,
+				"another ulogd already running with pid %d\n",
+				pid);
+			fclose(fp);
+			close(fd);
+			return -1;
+		}
+
+		close(fd);
+		fclose(fp);
+		unlink(ulogd_pidfile);
+
+		fd = open(ulogd_pidfile, O_RDWR | O_CREAT | O_EXCL, 0644);
+
+		if (fd < 0) {
 			ulogd_log(ULOGD_ERROR,
-				"unable to open configfile: %s\n",
-				ulogd_configfile);
-			break;
-		case -ERRMAND:
-			ulogd_log(ULOGD_ERROR,
-				"mandatory option \"%s\" not found\n",
-				config_errce->key);
-			break;
-		case -ERRMULT:
-			ulogd_log(ULOGD_ERROR,
-				"option \"%s\" occurred more than once\n",
-				config_errce->key);
-			break;
-		case -ERRUNKN:
-			ulogd_log(ULOGD_ERROR,
-				"unknown config key \"%s\"\n",
-				config_errce->key);
-			break;
-		case -ERRSECTION:
-			ulogd_log(ULOGD_ERROR,
-				"section \"%s\" not found\n", section);
-			break;
-		case -ERRTOOLONG:
-			if (config_errce->key)
-				ulogd_log(ULOGD_ERROR,
-					  "string value too long for key \"%s\"\n",
-					  config_errce->key);
-			else
-				ulogd_log(ULOGD_ERROR,
-					  "string value is too long\n");
-			break;
+				"cannot open %s (2nd time round): %d\n",
+				ulogd_pidfile, errno);
+			return -1;
+		}
 	}
-	return 1;
+
+	if (lock_fd(fd, 0) < 0) {
+		ulogd_log(ULOGD_ERROR, "cannot lock %s: %s\n", ulogd_pidfile,
+				strerror(errno));
+		close(fd);
+		return -1;
+	}
+	ulogd_pidfile_fd = fd;
+	return 0;
+}
+
+static int write_pidfile(int daemonize)
+{
+	FILE *fp;
+	if (!ulogd_pidfile)
+		return 0;
+
+	if (ulogd_pidfile_fd == -1) {
+		ulogd_log(ULOGD_ERROR, "unset pid file fd\n");
+		return -1;
+	}
+
+	if (daemonize) {
+		/* relocking as lock is not inherited */
+		if (lock_fd(ulogd_pidfile_fd, 1) < 0) {
+			ulogd_log(ULOGD_ERROR, "cannot lock %s: %d\n", ulogd_pidfile,
+					errno);
+			close(ulogd_pidfile_fd);
+			return -1;
+		}
+	}
+
+	fp = fdopen(ulogd_pidfile_fd, "w");
+	if (fp == NULL) {
+		ulogd_log(ULOGD_ERROR, "cannot fdopen %s: %d\n", ulogd_pidfile,
+				errno);
+		close(ulogd_pidfile_fd);
+		return -1;
+	}
+
+	fprintf(fp, "%d\n", getpid());
+	fflush(fp);
+
+	if (ftruncate(fileno(fp), ftell(fp)) < 0)
+		ulogd_log(ULOGD_NOTICE, "cannot ftruncate %s: %d\n",
+				ulogd_pidfile, errno);
+
+	/*
+	 * We do NOT close fd, since we want to keep the lock. However, we don't
+	 * want to keep the file descriptor in case of an exec().
+	 */
+	fcntl(ulogd_pidfile_fd, F_SETFD, FD_CLOEXEC);
+
+	created_pidfile = 1;
+
+	return 0;
+}
+
+static void cleanup_pidfile()
+{
+	if (!ulogd_pidfile || !created_pidfile)
+		return;
+
+	if (unlink(ulogd_pidfile) != 0)
+		ulogd_log(ULOGD_ERROR, "PID file %s could not be deleted: %d\n",
+				ulogd_pidfile, errno);
 }
 
 static void deliver_signal_pluginstances(int signal)
@@ -1022,13 +1348,20 @@ static void stop_pluginstances()
 
 	llist_for_each_entry(stack, &ulogd_pi_stacks, stack_list) {
 		llist_for_each_entry_safe(pi, npi, &stack->list, list) {
-			if ((pi->plugin->priv_size > 0 || *pi->plugin->stop) &&
-			    pluginstance_stop(pi)) {
+			if (*pi->plugin->stop && pluginstance_stop(pi)) {
 				ulogd_log(ULOGD_DEBUG, "calling stop for %s\n",
 					  pi->plugin->name);
 				(*pi->plugin->stop)(pi);
-				pi->private[0] = 0;
 			}
+
+			/* NB: plugin->stop() might access other plugin instances,
+			 * so we cannot free right away.
+			 */
+		}
+	}
+
+	llist_for_each_entry(stack, &ulogd_pi_stacks, stack_list) {
+		llist_for_each_entry_safe(pi, npi, &stack->list, list) {
 			free(pi);
 		}
 	}
@@ -1055,7 +1388,19 @@ static void stop_stack()
 }
 
 
-static void sigterm_handler(int signal)
+/* This is a real SIGTERM, SIGINT, SIGHUP, SIGALRM, SIGUSR1, SIGUSR2 handler */
+static void signal_handler(int signal)
+{
+	unsigned char	c = (unsigned char)signal;
+
+	(void)write(SIGNAL_TX_FD, &c, 1);
+}
+
+/* This is NOT a real signal handler.
+ * It is called in ulogd_main_loop() to avoid
+ * calling Async-Signal-UnSafe functions in signal handler.
+ */
+static void sigterm_handler_task(int signal)
 {
 
 	ulogd_log(ULOGD_NOTICE, "Terminal signal received, exiting\n");
@@ -1070,7 +1415,7 @@ static void sigterm_handler(int signal)
 	unload_plugins();
 #endif
 
-	if (logfile != NULL  && logfile != stdout && logfile != &syslog_dummy) {
+	if (logfile != NULL  && logfile != stdout) {
 		fclose(logfile);
 		logfile = NULL;
 	}
@@ -1080,24 +1425,31 @@ static void sigterm_handler(int signal)
 
 	config_stop();
 
+	cleanup_pidfile();
+
 	exit(0);
 }
 
-static void signal_handler(int signal)
+
+/* This is NOT a real signal handler.
+ * It is called in ulogd_main_loop() to avoid
+ * calling Async-Signal-UnSafe functions in signal handler.
+ */
+static void signal_handler_task(int signal)
 {
 	ulogd_log(ULOGD_NOTICE, "signal received, calling pluginstances\n");
 	
 	switch (signal) {
 	case SIGHUP:
 		/* reopen logfile */
-		if (logfile != stdout && logfile != &syslog_dummy) {
+		if (logfile != stdout && logfile != syslog_dummy) {
 			fclose(logfile);
 			logfile = fopen(ulogd_logfile, "a");
  			if (!logfile) {
 				fprintf(stderr, 
 					"ERROR: can't open logfile %s: %s\n", 
 					ulogd_logfile, strerror(errno));
-				sigterm_handler(signal);
+				sigterm_handler_task(signal);
 			}
 	
 		}
@@ -1109,6 +1461,19 @@ static void signal_handler(int signal)
 	deliver_signal_pluginstances(signal);
 }
 
+static void set_scheduler(void)
+{
+	struct sched_param schedparam;
+	int sched_type;
+
+	schedparam.sched_priority = sched_get_priority_max(SCHED_RR);
+	sched_type = SCHED_RR;
+
+	if (sched_setscheduler(0, sched_type, &schedparam) < 0)
+		fprintf(stderr, "WARNING: scheduler configuration failed:"
+			" %s\n", strerror(errno));
+}
+
 static void print_usage(void)
 {
 	printf("ulogd Version %s\n", VERSION);
@@ -1121,6 +1486,7 @@ static void print_usage(void)
 	printf("\t-v --verbose\tOutput info on standard output\n");
 	printf("\t-l --loglevel\tSet log level\n");
 	printf("\t-c --configfile\tUse alternative Configfile\n");
+	printf("\t-p --pidfile\tRecord ulogd PID in file\n");
 	printf("\t-u --uid\tChange UID/GID\n");
 	printf("\t-i --info\tDisplay infos about plugin\n");
 }
@@ -1134,6 +1500,7 @@ static struct option opts[] = {
 	{ "info", 1, NULL, 'i' },
 	{ "verbose", 0, NULL, 'v' },
 	{ "loglevel", 1, NULL, 'l' },
+	{ "pidfile", 1, NULL, 'p' },
 	{NULL, 0, NULL, 0}
 };
 
@@ -1150,7 +1517,7 @@ int main(int argc, char* argv[])
 
 	ulogd_logfile = strdup(ULOGD_LOGFILE_DEFAULT);
 
-	while ((argch = getopt_long(argc, argv, "c:dvl:h::Vu:i:", opts, NULL)) != -1) {
+	while ((argch = getopt_long(argc, argv, "c:p:dvl:h::Vu:i:", opts, NULL)) != -1) {
 		switch (argch) {
 		default:
 		case '?':
@@ -1179,6 +1546,9 @@ int main(int argc, char* argv[])
 		case 'c':
 			ulogd_configfile = optarg;
 			break;
+		case 'p':
+			ulogd_pidfile = optarg;
+			break;
 		case 'u':
 			change_uid = 1;
 			user = strdup(optarg);
@@ -1206,9 +1576,15 @@ int main(int argc, char* argv[])
 	}
 
 	/* command line has precedence on config file */
-	if (loglevel)
+	if (loglevel) {
 		loglevel_ce.u.value = loglevel;
 		loglevel_ce.flag |= CONFIG_FLAG_VAL_PROTECTED;
+	}
+
+	if (ulogd_pidfile) {
+		if (create_pidfile() < 0)
+			warn_and_exit(0);
+	}
 
 	if (daemonize && verbose) {
 		verbose = 0;
@@ -1217,6 +1593,19 @@ int main(int argc, char* argv[])
 			  " with daemon mode).\n");
 	}
 
+	if (daemonize){
+		if (daemon(0, 0) < 0) {
+			ulogd_log(ULOGD_FATAL, "can't daemonize: %s (%d)\n",
+				  strerror(errno), errno);
+			warn_and_exit(daemonize);
+		}
+	}
+
+	if (ulogd_pidfile) {
+		if (write_pidfile(daemonize) < 0)
+			warn_and_exit(0);
+	}
+
 	if (config_register_file(ulogd_configfile)) {
 		ulogd_log(ULOGD_FATAL, "error registering configfile \"%s\"\n",
 			  ulogd_configfile);
@@ -1224,7 +1613,7 @@ int main(int argc, char* argv[])
 	}
 
 	/* parse config file */
-	if (parse_conffile("global", &ulogd_kset)) {
+	if (ulogd_parse_configfile("global", &ulogd_kset)) {
 		ulogd_log(ULOGD_FATAL, "unable to parse config file\n");
 		warn_and_exit(daemonize);
 	}
@@ -1235,6 +1624,13 @@ int main(int argc, char* argv[])
 		warn_and_exit(daemonize);
 	}
 
+	errno = 0;
+	if (nice(-1) == -1) {
+		if (errno != 0)
+			ulogd_log(ULOGD_ERROR, "Could not nice process: %s\n",
+				  strerror(errno));
+	}
+
 	if (change_uid) {
 		ulogd_log(ULOGD_NOTICE, "Changing UID / GID\n");
 		if (setgid(gid)) {
@@ -1261,31 +1657,18 @@ int main(int argc, char* argv[])
 		}
 	}
 
-	errno = 0;
-	if (nice(-1) == -1) {
-		if (errno != 0)
-			ulogd_log(ULOGD_ERROR, "Could not nice process: %s\n",
-				  strerror(errno));
+	if (create_signal_channel() < 0) {
+		ulogd_log(ULOGD_FATAL, "can't create signal channel\n");
+		warn_and_exit(daemonize);
 	}
 
-
-	if (daemonize){
-		if (fork()) {
-			exit(0);
-		}
-		if (logfile != stdout)
-			fclose(stdout);
-		fclose(stderr);
-		fclose(stdin);
-		setsid();
-	}
-
-	signal(SIGTERM, &sigterm_handler);
-	signal(SIGINT, &sigterm_handler);
+	signal(SIGTERM, &signal_handler);
+	signal(SIGINT, &signal_handler);
 	signal(SIGHUP, &signal_handler);
 	signal(SIGALRM, &signal_handler);
 	signal(SIGUSR1, &signal_handler);
 	signal(SIGUSR2, &signal_handler);
+	set_scheduler();
 
 	ulogd_log(ULOGD_INFO, 
 		  "initialization finished, entering main loop\n");
@@ -1293,6 +1676,6 @@ int main(int argc, char* argv[])
 	ulogd_main_loop();
 
 	/* hackish, but result is the same */
-	sigterm_handler(SIGTERM);	
+	sigterm_handler_task(SIGTERM);
 	return(0);
 }
